#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#AERONET_processor.py

#module which handles parsing of AERONET network data files.

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import crossnetwork_processing_functions
import dateutil
from dateutil.tz import gettz
tzinfos = {"GMT": gettz("GMT")}
import GHOST_standards
import numpy as np
import standardisation_mapping_functions
import time

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def read_data(slurm_job_ID, network, parameter, year):

    '''Convert daily provider original files across 1 year into GHOST standard .npy station specific annual files'''

    #initialise log file string
    log_file_str = ''

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = crossnetwork_processing_functions.get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get specific data files for current year
    data_files = crossnetwork_processing_functions.get_data_files(run_details, parameter_details, specific_year=year)
    
    #define select variable fields (also define date/time variable names)
    if parameter_details['aeronet_matrix'] == 'oneill':
        date_var = 'Date_(dd:mm:yyyy)'
        time_var = 'Time_(hh:mm:ss)'
        req_vars = ['AERONET_Site',date_var,time_var,parameter_details['aeronet_parameter_name'],'Fine_Mode_AOD_500nm[tau_f]','Data_Quality_Level','AERONET_Instrument_Number','AERONET_Site_Name','Site_Latitude(Degrees)','Site_Longitude(Degrees)','Site_Elevation(m)','contact_name','contact_institution','contact_email_address']
    elif parameter_details['aeronet_matrix'] == 'directsun':   
        date_var = 'Date(dd:mm:yyyy)'
        time_var = 'Time(hh:mm:ss)'
        if parameter == 'AOD_550nm':
            req_vars = ['AERONET_Site',date_var,time_var,'AOD_440nm','AOD_675nm','AOD_870nm','440-870_Angstrom_Exponent','Data_Quality_Level','AERONET_Instrument_Number','AERONET_Site_Name','Site_Latitude(Degrees)','Site_Longitude(Degrees)','Site_Elevation(m)','contact_name','contact_institution','contact_email_address']
        else:
            req_vars = ['AERONET_Site',date_var,time_var,parameter_details['aeronet_parameter_name'],'Data_Quality_Level','AERONET_Instrument_Number','AERONET_Site_Name','Site_Latitude(Degrees)','Site_Longitude(Degrees)','Site_Elevation(m)','contact_name','contact_institution','contact_email_address']

    #split data files into different lists by data level
    lev10_data_files = []
    lev15_data_files = []
    lev20_data_files = []
    for data_file in data_files:
        if '/lev1.0/' in data_file:
            lev10_data_files.append(data_file)
        elif '/lev1.5/' in data_file:
            lev15_data_files.append(data_file)
        elif '/lev2.0/' in data_file:
            lev20_data_files.append(data_file)

    #iterate through each set of specific data level files
    for data_level, data_files in zip(['1.0','1.5','2.0'], [lev10_data_files, lev15_data_files, lev20_data_files]):
        log_file_str += 'PROCESSING DATA LEVEL: {}\n'.format(data_level) 

        #define dictionary for storing all parsed data (per specific data level)
        all_data = {k:[] for k in req_vars}
    
        #iterate through daily data files (for specific data level)
        for data_file in data_files:
            log_file_str += 'PROCESSING FILE: {}\n'.format(data_file) 

            #create object for parsing AERONET daily file
            parsing_obj = crossnetwork_processing_functions.parse_AERONET_file()

            #open and read file contents into memory
            with open(data_file, "r") as f:
                page = f.read()

            #parse read file contents 
            parsing_obj.feed(page)

            #iterate through read object dictionary keys
            for var in list(parsing_obj.var_data.keys()):
                #is read object key one of required vars?
                if var in req_vars:
                    #if so --> append to all data dictionary
                    all_data[var] = np.append(all_data[var], parsing_obj.var_data[var])

            #add parsed contact information to all_data dictionary
            all_data['contact_name'] = np.append(all_data['contact_name'], [parsing_obj.contact_name]*len(parsing_obj.var_data['AERONET_Site'])) 
            all_data['contact_institution'] = np.append(all_data['contact_institution'], [parsing_obj.contact_institution]*len(parsing_obj.var_data['AERONET_Site'])) 
            all_data['contact_email_address'] = np.append(all_data['contact_email_address'], [parsing_obj.contact_email_address]*len(parsing_obj.var_data['AERONET_Site'])) 

        #get all unique "AERONET_Site" variables --> take these to be the unique station references
        unique_station_references = np.unique(all_data['AERONET_Site'])

        #get current AERONET data level (period removed)
        data_level_mod = data_level.replace(".","") 
        #get mask of all data of correct data level (for sanity)
        data_level_bool = np.array([True if var == data_level_mod else False for var in all_data['Data_Quality_Level']])

        #get dictionary necessary for mapping of AERONET file data flags to standard data flag codes
        data_flag_mapping_dictionary = standardisation_mapping_functions.network_data_flag_to_standard_data_flag_mapping['AERONET']

        #iterate through unique stations references in data
        for orig_station_reference in unique_station_references:

            #set flag to know when have written data for station 
            written_data = False

            #strip underscore from original unique station reference string
            station_reference = orig_station_reference.replace('_','')

            log_file_str += '{} -- PARSING RAW DATA\n'.format(station_reference)

            #create site instances of metadata/data dictionaries
            site_data =     GHOST_standards.get_standard_data(parameter_details)
            site_metadata = GHOST_standards.get_standard_metadata(parameter_details)

            #isolate variable data specific for station
            unique_station_bool = all_data['AERONET_Site'] == orig_station_reference
            #further isolate variable data to be only that associated with a data quality level that is currently being processed 
            unique_station_bool = data_level_bool & unique_station_bool
            station_var_data = {req_var:all_data[req_var][unique_station_bool] for req_var in req_vars}
            if parameter == 'AOD_550nm':
                station_var_data['AOD_550nm'] = []

            #------------------------------------------------------------------------#
            #PARSE TIME/DATA
            #------------------------------------------------------------------------#
        
            #parse datetime for each reported timepoint 
            #first create a unified date/time string (yyyy-mm-dd hh:mm:ss Z, of the measurement --> z is always GMT)
            #secondly parse into datetime (taking into account timezone information). then convert to UTC datetime, then round to nearest minute    
            start_dt = [crossnetwork_processing_functions.round_to_minute(dateutil.parser.parse('{}-{}-{} {} GMT'.format(date[6:],date[3:5],date[:2],time), tzinfos=tzinfos).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date,time in zip(station_var_data[date_var],station_var_data[time_var])] 
            #convert each start measurement datetime to minutes from reference time
            frac_time = [crossnetwork_processing_functions.datetime_to_fractional_minutes_from_reference(dt) for dt in start_dt]

            #------------------------------------------------------------------------#
            #PARSE METADATA
            #------------------------------------------------------------------------#

            #GHOST version
            site_metadata['GHOST_version']['value'] = [run_details['ghost_version']]*len(frac_time)

            #initialise string which will store all GHOST process warnings as empty            
            site_metadata['process_warnings']['value'] = ['']*len(frac_time)

            #check if measurement position metadata is suspect and write it to metadata/data dictionaries
            for meta_ii in range(len(frac_time)):
                meas_var_dict = crossnetwork_processing_functions.check_measurement_position(longitude=station_var_data['Site_Longitude(Degrees)'][meta_ii], latitude=station_var_data['Site_Latitude(Degrees)'][meta_ii], measurement_altitude=station_var_data['Site_Elevation(m)'][meta_ii])
                site_metadata['longitude']['value'].append(meas_var_dict['longitude'])
                site_metadata['latitude']['value'].append(meas_var_dict['latitude'])
                site_metadata['measurement_altitude']['value'].append(meas_var_dict['measurement_altitude'])

            #save meta_update_stamp, set as copy of time array
            site_metadata['meta_update_stamp']['value'] = frac_time

            #write known metadata  
            site_metadata['network']['value'] = [network]*len(frac_time)
            site_metadata['station_name']['value'] = station_var_data['AERONET_Site_Name']
            site_metadata['data_level']['value'] = [data_level]*len(frac_time)
            site_metadata['contact_name']['value'] = station_var_data['contact_name']
            site_metadata['contact_institution']['value'] = station_var_data['contact_institution']
            site_metadata['contact_email_address']['value'] = station_var_data['contact_email_address']

            #process measurement information
            site_metadata, measurement_qa_flags, standard_method_abbreviation = standardisation_mapping_functions.map_network_measurement_information_to_standard_measurement_information(parameter_details['aeronet_matrix'], site_metadata, run_details)

            #append station_reference now that we have the standard method abbreviation
            site_metadata['station_reference']['value'] = ['{}_{}'.format(station_reference, standard_method_abbreviation)]*len(frac_time)

            #iterate through metadata variables, if any columns have not been appended to add a NaN
            for column_header in list(site_metadata.keys()):
                n_column = len(site_metadata[column_header]['value'])
                #if column is empty, then append NaN
                if n_column < 1:
                    site_metadata[column_header]['value'] = [np.NaN] * len(frac_time)

            #get data file download timestamp
            site_metadata['data_download_stamp']['value'] = [crossnetwork_processing_functions.get_datafile_download_timestamp(data_file)] * len(frac_time)

            #------------------------------------------------------------------------#
            #PARSE FLAGS/QA
            #------------------------------------------------------------------------#

            #create empty list for standard data flags associated with each measurement
            flags = []
            qa = []
 
            #process data flags (artificially created)
            for val_ii, val in enumerate(station_var_data[date_var]):
 
                #create empty list for standard data flags associated with each measurement
                standard_data_flag_names = []

                #set GHOST/network decreed validity as initially 'V'
                GHOST_decreed_validity = 'V'
                network_decreed_validity = 'V'
 
                #for AERONET O'Neill matrix measurements, artificially create QA flag variable where FINE_MODE_FRAC_500nm is outside valid bounds
                if parameter_details['aeronet_matrix'] == 'oneill':
 
                    #is FINE_MODE_FRAC_500nm outside valid bounds?
                    float_val = float(station_var_data['Fine_Mode_AOD_500nm[tau_f]'][val_ii])
                    if (float_val < GHOST_standards.standard_parameters['FINE_MODE_FRAC_500nm']['extreme_lower_limit']) or (float_val > GHOST_standards.standard_parameters['FINE_MODE_FRAC_500nm']['extreme_upper_limit']):
                        data_flag = 'FM500FRAC_I' 
                        #get mapped standard data flag specific dictionary 
                        mapped_standard_data_flag_dict = data_flag_mapping_dictionary[data_flag]
                        #unpack mapped standard data flag key variables
                        standard_data_flag_names.append(mapped_standard_data_flag_dict['standard_data_flag_name'])
                        GHOST_decreed_validity = mapped_standard_data_flag_dict['GHOST_decreed_validity']
                        network_decreed_validity = mapped_standard_data_flag_dict['network_decreed_validity']

                #when calculating AOD 550nm from AERONET directsun measurements, artificially create QA flag variable where one of used components is outside valid bounds
                elif (parameter_details['aeronet_matrix'] == 'directsun') & (parameter == 'AOD_550nm'): 
                
                    #Is AOD_440nm, AOD675nm, AOD870nm or AE_440-870nm outside valid bounds?
                    #calculate AOD_550nm (if all needed component values in valid ranges)
                    float_aod440 = float(station_var_data['AOD_440nm'][val_ii]) 
                    float_aod675 = float(station_var_data['AOD_675nm'][val_ii]) 
                    float_aod870 = float(station_var_data['AOD_870nm'][val_ii]) 
                    float_ae440_870 = float(station_var_data['440-870_Angstrom_Exponent'][val_ii]) 
                    aod440_valid = False
                    aod675_valid = False
                    aod870_valid = False
                    ae440_870_valid = False

                    #AE_440-870nm outside valid bounds?
                    if (float_ae440_870 >= GHOST_standards.standard_parameters['AE_440-870nm']['extreme_lower_limit']) & (float_ae440_870 <= GHOST_standards.standard_parameters['AE_440-870nm']['extreme_upper_limit']):
                        ae440_870_valid = True
                        #AOD_440nm outside valid bounds?
                        if (float_aod440 >= GHOST_standards.standard_parameters['AOD_440nm']['extreme_lower_limit']) & (float_aod440 <= GHOST_standards.standard_parameters['AOD_440nm']['extreme_upper_limit']):
                            aod_550_adjust1 = (float_aod440/(440**(-float_ae440_870)))*(550**(-float_ae440_870))
                            aod440_valid = True
                        #AOD_675nm outside valid bounds?
                        if (float_aod675 >= GHOST_standards.standard_parameters['AOD_675nm']['extreme_lower_limit']) & (float_aod675 <= GHOST_standards.standard_parameters['AOD_675nm']['extreme_upper_limit']):
                            aod_550_adjust2 = (float_aod675/(675**(-float_ae440_870)))*(550**(-float_ae440_870))
                            aod675_valid = True
                        #AOD_870nm outside valid bounds?
                        if (float_aod870 >= GHOST_standards.standard_parameters['AOD_870nm']['extreme_lower_limit']) & (float_aod870 <= GHOST_standards.standard_parameters['AOD_870nm']['extreme_upper_limit']):   
                            aod_550_adjust3 = (float_aod870/(870**(-float_ae440_870)))*(550**(-float_ae440_870))
                            aod870_valid = True

                    #if one of needed values is outside of valid bounds then append data flag, and also set AOD 550nm value as NaN
                    if (aod440_valid == False) or (aod675_valid == False) or (aod870_valid == False) or (ae440_870_valid == False):
                        data_flag = 'AOD550CON_I' 
                        #get mapped standard data flag specific dictionary 
                        mapped_standard_data_flag_dict = data_flag_mapping_dictionary[data_flag]
                        #unpack mapped standard data flag key variables
                        standard_data_flag_names.append(mapped_standard_data_flag_dict['standard_data_flag_name'])
                        GHOST_decreed_validity = mapped_standard_data_flag_dict['GHOST_decreed_validity']
                        network_decreed_validity = mapped_standard_data_flag_dict['network_decreed_validity']
                        station_var_data['AOD_550nm'].append(np.NaN)
                    #append valid calculated AOD 500nm
                    else:
                        station_var_data['AOD_550nm'].append(np.mean([aod_550_adjust1,aod_550_adjust2,aod_550_adjust3]))
 
                #get standard flag codes associated with each standard data flag name, and append to list to write out
                standard_data_flag_codes = [GHOST_standards.standard_data_flag_name_to_data_flag_code[standard_data_flag_name] for standard_data_flag_name in standard_data_flag_names]
 
                #process qa information associated with flags
                qa_codes = []
                #append qa flag code if standard data flag is GHOST decreed invalid
                if GHOST_decreed_validity == 'I':
                    qa_codes.append(GHOST_standards.standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
                #append qa flag if standard data flag is network decreed invalid
                if network_decreed_validity == 'I':
                    qa_codes.append(GHOST_standards.standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])
            
                #iterate through measurement QA flags and write standard qa flag code if required
                for standard_flag_name in list(measurement_qa_flags.keys()):
                    if measurement_qa_flags[standard_flag_name] == True:
                        qa_codes.append(GHOST_standards.standard_QA_name_to_QA_code[standard_flag_name])

                #append flags/qa codes to list to write out
                flags.append(standard_data_flag_codes)
                qa.append(qa_codes)

            #----------------------------------------------------------------------------#
            #WRITE OUT DATA
            #----------------------------------------------------------------------------#
            site_data['time']['value'] = frac_time
            site_data[parameter_details['bsc_parameter_name']]['value'] = station_var_data[parameter_details['aeronet_parameter_name']]
            site_data['temporal_resolution']['value'] = [0]*len(frac_time)   
            site_data['flag']['value'] = flags
            site_data['qa']['value'] = qa
            #iterate through data variables, if any columns have not been appended to, add a list of NaNs (equal to length of time array)
            for column_header in list(site_data.keys()):
                n_column = len(site_data[column_header]['value'])
                #if column is empty, then append NaN
                if n_column == 0:
                     if site_data[column_header]['data_type'] != np.uint8:
                         site_data[column_header]['value'] = [np.NaN]*len(site_data['time']['value'])
                     else:
                         site_data[column_header]['value'] = [255]*len(site_data['time']['value'])
            #------------------------------------------------------
 
            #standardise metadata/data and write out specific file station data to temporary .npy
            crossnetwork_processing_functions.post_process_file_parse(site_metadata,site_data,run_details,parameter_details)

    #write a valid process log file
    log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
    crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,0,log_file_str,year)
